import warnings

import bilby
import numpy as np
from bilby.core.prior import DeltaFunction, Normal, PriorDict, Uniform
from jax.random import PRNGKey
from starccato_jax import StarccatoVAE
from starccato_jax.data import load_training_data

# pip install scipy==1.14.0


vae = StarccatoVAE()
train_dat, val_dat = load_training_data()

rng = PRNGKey(0)

warnings.filterwarnings("ignore", "Wswiglal-redir-stdio")

# Set the duration and sampling frequency of the data segment that we're going
# to inject the signal into.
# These are fixed by the resolution in the injection file that we are using.

sampling_frequency = 4096
n_timestamps = 256
duration = n_timestamps / sampling_frequency
time = np.linspace(0, duration, n_timestamps)
t0 = time[53]  # This is the time of the core-bounce

LUMIN_DIST = 10  # kpc
GEOCENT_TIME = 1126259642.413

# Specify the output directory and the name of the simulation.
outdir = "outdir"
label = "supernova"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

np.random.seed(170801)


def get_latent_vector_prior():
    priors = {}
    for i in range(vae.latent_dim):
        priors[f"z_{i}"] = Normal(0, 1, name=f"z_{i}")
    return priors


prior = PriorDict(
    dict(
        **get_latent_vector_prior(),
        luminosity_distance=Uniform(2, 20, "luminosity_distance", unit="kpc"),
        ra=Uniform(0, 2 * np.pi, "ra", boundary="periodic"),
        dec=Uniform(-np.pi / 2, np.pi / 2, "dec", boundary="periodic"),
        geocent_time=Uniform(
            GEOCENT_TIME - 1, GEOCENT_TIME + 1, "geocent_time", unit="s"
        ),
        psi=DeltaFunction(name="psi", unit="radian", peak=0.0),
    ),
)

injection_parameters = dict(
    luminosity_distance=LUMIN_DIST,  # kpc
    geocent_time=GEOCENT_TIME,
    ra=0,
    dec=0,
    psi=0,
    **PriorDict(get_latent_vector_prior()).sample(1),
)


def supernova(freq_array, luminosity_distance, z, **kwargs):
    """
    A source model that reads a simulation from a text file.

    This was originally intended for use with supernova simulations, but can
    be applied to any source class.

    Parameters
    ----------
    frequency_array: array-like
        Unused (but required by the source model interface)

    Returns
    -------
    dict:
        A dictionary containing the plus and cross components of the signal.
    """
    # print all args
    rng = kwargs["rng"]
    # print(f"z: {z}, {z.shape}")
    waveform = vae.generate(z=z, rng=rng)[0]
    # print(f"waveform: {waveform}, {waveform.shape}")

    # waveforms generated at 10kpc, so scale to the luminosity distance
    scaling = 1e-21 * (10.0 / luminosity_distance)

    # idk if the signal is hcross/hplus, so we just duplicate for now... #TODO
    h_plus = scaling * waveform
    return {"plus": h_plus, "cross": h_plus}


def parameter_conversion(parameters):
    new_params = {}
    new_params["luminosity_distance"] = parameters["luminosity_distance"]
    z = np.array([parameters[f"z_{i}"] for i in range(vae.latent_dim)])
    # ennsure z is of shape (1, latent_dim)
    if z.shape != (1, vae.latent_dim):
        z = z.reshape(1, vae.latent_dim)

    new_params["z"] = z

    return new_params, parameters


# Create the waveform_generator using a supernova source function
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    time_domain_source_model=supernova,
    parameters=injection_parameters,
    parameter_conversion=parameter_conversion,
    waveform_arguments=dict(rng=rng),
)

# Set up interferometers (LIGO-Hanford (H1) and LIGO-Livingston (L1))
ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"],
)

injection_strain_time = waveform_generator.time_domain_strain(
    injection_parameters
)
injection_strain = waveform_generator.frequency_domain_strain(
    injection_parameters
)

ifos.inject_signal(
    injection_polarizations=injection_strain,
    parameters=injection_parameters,
    raise_error=False,
)

import matplotlib.pyplot as plt
from bilby.gw import utils as gwutils

DATA_COL = "tab:gray"
SIGNAL_COL = "tab:orange"
PSD_COL = "black"


def plot_freq_domain(ifo: bilby.gw.detector.Interferometer, freq_signal):
    fig, ax = plt.subplots()
    df = (
        ifo.strain_data.frequency_array[1] - ifo.strain_data.frequency_array[0]
    )
    asd = gwutils.asd_from_freq_series(
        freq_data=ifo.strain_data.frequency_domain_strain, df=df
    )

    ax.loglog(
        ifo.strain_data.frequency_array[ifo.strain_data.frequency_mask],
        asd[ifo.strain_data.frequency_mask],
        color=DATA_COL,
        label=f"{ifo.name} Data",
        alpha=0.5,
        lw=3,
    )
    ax.loglog(
        ifo.strain_data.frequency_array[ifo.strain_data.frequency_mask],
        ifo.amplitude_spectral_density_array[ifo.strain_data.frequency_mask],
        color=PSD_COL,
        lw=1.0,
        label=ifo.name + " ASD",
    )

    signal_asd = gwutils.asd_from_freq_series(freq_data=freq_signal, df=df)

    ax.loglog(
        ifo.strain_data.frequency_array[ifo.strain_data.frequency_mask],
        signal_asd[ifo.strain_data.frequency_mask],
        color=SIGNAL_COL,
        label=f'Signal (SNR: {ifo.meta_data["optimal_SNR"]:.2f})',
    )
    ax.grid(True)
    ax.set_ylabel(r"Strain [strain/$\sqrt{\rm Hz}$]")
    ax.set_xlabel(r"Frequency [Hz]")
    ax.legend(loc="best")


def plot_time_domain(ifo, time_signal):
    fig, ax = plt.subplots()
    strain = ifo.strain_data.time_domain_strain
    t0 = ifo.strain_data.start_time
    x = ifo.strain_data.time_array - t0
    xlabel = f"GPS time [s] - {t0}"
    # unroll raw data
    strain = np.roll(strain, 55)
    ax.plot(
        x, strain, color=DATA_COL, label=f"{ifo.name} Data", alpha=0.5, lw=3
    )
    ax.plot(
        x,
        time_signal,
        color=SIGNAL_COL,
        label=f'Signal (SNR: {ifo.meta_data["optimal_SNR"]:.2f})',
    )
    ax.set_xlabel(xlabel)
    ax.set_ylabel("Strain")
    ax.legend()
    fig.tight_layout()
    return ax


plot_freq_domain(ifos[0], injection_strain["plus"])
plot_time_domain(ifos[0], injection_strain_time["plus"])

signals = []
for i in range(100):
    sample = prior.sample()
    s = waveform_generator.time_domain_strain(sample)["plus"]
    signals.append(s)

signal_ci = np.quantile(np.array(signals), [0.05, 0.5, 0.95], axis=0)

t0 = ifos[0].strain_data.start_time
x = ifos[0].strain_data.time_array - t0
ax = plot_time_domain(ifos[0], injection_strain_time["plus"])
ax.fill_between(
    x, signal_ci[0], signal_ci[2], alpha=0.25, color="tab:blue", label="Prior"
)
ax.legend()

# Run sampler.
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)

result = bilby.run_sampler(
    likelihood=likelihood,
    priors=prior,
    sampler="emcee",
    outdir=outdir,
    label=label,
    clean=True,
)



import glob
import os

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
from tqdm.auto import tqdm

__all__ = ["pp_test"]


def _load_results_and_compute_credible_levels(
    result_regex: str, dim: int = 32
) -> np.ndarray:
    """
    Load posterior samples from NetCDF files, compute credible levels for injected parameters,
    and return an array of credible levels.

    The credible level for a true latent parameter value is computed as the fraction of posterior
    samples that are smaller than the true value. This provides an empirical cumulative
    distribution function (ECDF)-based measure of how well the posterior distribution
    recovers the true injected parameters.

    Parameters:
    result_regex (str): A regex pattern to locate result files.
    dim (int): The number of latent dimensions in the data.

    Returns:
    np.ndarray: A (dim, N) array containing the credible levels for each parameter and each file.
    """
    result_files = glob.glob(result_regex)
    assert (
        len(result_files) > 0
    ), f"No files found with pattern {result_regex}."
    credible_levels = np.zeros((dim, len(result_files)))
    for ri, result_file in tqdm(
        enumerate(result_files), total=len(result_files)
    ):
        try:
            idata = az.from_netcdf(result_file)
            zpost = idata.posterior["z"].stack(sample=("chain", "draw")).values
            ztrue = idata.sample_stats["true_latent"].to_numpy().ravel()
            for i, _ztrue in enumerate(ztrue):
                posterior_samples = zpost[i].flatten()
                credible_levels[i, ri] = np.mean(posterior_samples < _ztrue)
        except Exception as e:
            print(f"Error processing {result_file}: {e}")
            credible_levels[:, ri] = np.nan
    return credible_levels


def _make_pp_plot(
    credible_levels: np.ndarray,
    filename="pp.png",
    confidence_interval=[0.68, 0.95, 0.997],
    title=True,
    color="tab:orange",
    **kwargs,
):
    """
    Generate a probability-probability (P-P) plot comparing empirical credible levels
    to expected uniform behavior.

    The plot is a diagnostic tool for checking whether the posterior credible levels
    are uniformly distributed, which is expected if the inference is well-calibrated.

    Confidence intervals are computed using a binomial cumulative distribution function.
    Each credible level distribution is tested against a uniform distribution using
    the Kolmogorov-Smirnov test.

    Parameters:
    credible_levels (np.ndarray): An array of credible levels from the posterior samples.
    filename (str): The output filename for the P-P plot.
    confidence_interval (list of float): Confidence intervals for shading.
    title (bool): Whether to include a title on the plot.
    confidence_interval_alpha (float): Alpha value for confidence shading.

    Returns:
    Tuple[matplotlib.figure.Figure, dict]: A matplotlib figure and a dictionary of p-values.
    """
    x_values = np.linspace(0, 1, 1001)
    credible_levels = credible_levels[~np.isnan(credible_levels).any(axis=1)]
    dim, N = credible_levels.shape
    assert (
        N > dim
    ), f"Number of results {N} must be greater than the number of dim {dim}."

    fig, ax = plt.subplots(figsize=(3.5, 3.5))

    grays = plt.cm.Greys(np.linspace(0.2, 0.5, len(confidence_interval)))

    # Confidence interval shading
    for i, ci in enumerate(confidence_interval[::-1]):
        edge = (1 - ci) / 2
        lower = scipy.stats.binom.ppf(edge, N, x_values) / N
        upper = scipy.stats.binom.ppf(1 - edge, N, x_values) / N
        ax.fill_between(
            x_values, lower, upper, color=grays[i], lw=0, zorder=10
        )
        # draw line at theses values
        ax.plot(x_values, lower, color=grays[i], lw=1, zorder=10)
        ax.plot(x_values, upper, color=grays[i], lw=1, zorder=10)

    pvalues = []
    for i in range(len(credible_levels)):
        pp = np.array([(credible_levels[i] < x).mean() for x in x_values])
        pvalue = scipy.stats.kstest(credible_levels[i], "uniform").pvalue
        pvalues.append(pvalue)
        print(f"{i}: {pvalue}")
        ax.plot(
            x_values, pp, **kwargs, color=color, alpha=0.6, lw=0.3, zorder=110
        )

    combined_pvalue = scipy.stats.combine_pvalues(pvalues)[1]
    print(f"Combined p-value: {combined_pvalue}")

    if title:
        ax.set_title(f"N = {N}, p-value={combined_pvalue:.4f}")
    ax.set_xlabel("C.I.")
    ax.set_ylabel("Fraction of events in C.I.")
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    ax.legend(
        handles=[
            plt.Line2D([0], [0], color=grays[0], lw=2, label="68%"),
            plt.Line2D([0], [0], color=grays[1], lw=2, label="95%"),
            plt.Line2D([0], [0], color=grays[2], lw=2, label="99.7%"),
            plt.Line2D(
                [0], [0], color=color, lw=2, label=r"$\rm{CDF}_z(\rm{C.I.})$"
            ),
        ],
        loc="upper left",
        frameon=False,
    )
    fig.tight_layout()
    plt.savefig(filename, dpi=500)

    return fig, {"combined_pvalue": combined_pvalue, "pvalues": pvalues}


def pp_test(
    credible_levels_fpath: str,
    plot_fname: str,
    result_regex: str = None,
    include_title: bool = True,
    dim: int = 32,
):
    """
    Perform a P-P test by loading posterior samples, computing credible levels,
    and generating a P-P plot for model validation.

    The function first checks whether the credible levels have already been computed and saved.
    If not, it computes them from the result files. Then, it generates a P-P plot and saves it.

    Parameters:
    result_regex (str): Regex pattern to locate NetCDF result files.
    credible_levels_fpath (str): Path to save or load credible levels (must be .npy).
    plot_fname (str): Path to save the P-P plot (must be .png or .pdf).
    include_title (bool): Whether to include a title in the P-P plot.

    Returns:
    None
    """
    assert credible_levels_fpath.endswith(".npy")
    assert plot_fname.endswith(".png") or plot_fname.endswith(".pdf")

    if not os.path.exists(credible_levels_fpath):
        cred_level = _load_results_and_compute_credible_levels(
            result_regex=result_regex, dim=dim
        )
        np.save(credible_levels_fpath, cred_level)
    cred_level = np.load(credible_levels_fpath)
    _make_pp_plot(cred_level, filename=plot_fname, title=include_title)



import os

import jax
import pytest
from starccato_jax import StarccatoVAE

from starccato_sampler.sampler import sample

RNG = jax.random.PRNGKey(0)


@pytest.fixture
def injection():
    vae = StarccatoVAE()
    # true_z = jax.random.normal(RNG, (1, vae.latent_dim))
    true_z = jax.numpy.zeros((1, vae.latent_dim))
    true_signal = vae.generate(z=true_z)[0]
    return true_signal, true_z.ravel()


def test_sampler(injection, outdir):
    true_signal, true_z = injection
    sample(
        true_signal,
        outdir=outdir,
        num_warmup=50,
        num_samples=100,
        num_chains=1,
        verbose=True,
        stepping_stone_lnz=False,
        truth=dict(signal=true_signal, latent=true_z),
    )
    assert os.path.exists(outdir)
    assert os.path.exists(os.path.join(outdir, "inference.nc"))
    assert os.path.exists(os.path.join(outdir, "ci_plot.png"))
    assert os.path.exists(os.path.join(outdir, "1d_marginals.png"))



import jax.numpy as jnp
import numpy as np
from scipy.stats import beta as BetaDist


def beta_spaced_samples(
    n: int, alpha: float, beta: float, min: float = 1e-8, max: float = 1
):
    """Generate N samples spaced according to a Beta(a, b) distribution."""
    linspace = jnp.linspace(min, max, n)  # Uniformly spaced values
    return jnp.array(BetaDist.ppf(linspace, alpha, beta))



import os
from typing import List

import arviz as az
import corner
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from starccato_jax import StarccatoVAE
from starccato_jax.credible_intervals import coverage_probability, pointwise_ci
from starccato_jax.plotting import add_quantiles


def sampler_diagnostic_plots(inf_object: az.InferenceData, outdir: str):
    plot_trace(inf_object, outdir)
    plot_1d_marginals(inf_object, outdir)


def plot_trace(inf_object: az.InferenceData, outdir: str):
    # Plot the trace plot
    axes = az.plot_trace(inf_object, var_names=["z"])
    zdim = inf_object.posterior["z"].shape[-1]

    if "true_latent" in inf_object.sample_stats:
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        true_z = inf_object.sample_stats["true_latent"].to_numpy().ravel()
        for i, _z in enumerate(true_z):
            axes[0, 0].axvline(
                _z, color=colors[i % len(colors)], ls="--", alpha=0.3
            )

    axes[0, 0].set_ylim(bottom=0)
    plt.savefig(os.path.join(outdir, "trace_plot.png"))
    plt.close()


def plot_corner(inf_object: az.InferenceData, outdir: str):
    corner.corner(inf_object, var_names=["z"])
    plt.savefig(os.path.join(outdir, "corner_plot.png"))
    plt.close()


def plot_1d_marginals(
    inf_object: az.InferenceData, outdir: str, color="lightblue"
):
    zdim = inf_object.posterior["z"].shape[-1]
    zpost = inf_object.posterior["z"].stack(sample=("chain", "draw")).values.T
    figsize = (3, zdim * 0.25)
    # plot a vertical violin plot of the zpost
    fig, ax = plt.subplots(figsize=figsize)
    boxplot = ax.boxplot(
        zpost,
        vert=False,
        patch_artist=True,
        showfliers=False,
        capwidths=0,
        widths=0.1,
        medianprops=dict(
            marker="s",
            markersize=4,
            markerfacecolor=color,
            markeredgecolor=color,
            lw=0,
        ),
        boxprops=dict(facecolor=color, edgecolor=color),
        whiskerprops=dict(color=color),
        showmeans=False,
    )
    ax.set_ylabel("Z idx")
    ax.set_xlabel("Z value")
    if "true_latent" in inf_object.sample_stats:
        true_z = inf_object.sample_stats["true_latent"].to_numpy().ravel()
        # draw a short line at the true z value for each box
        box_height = (
            boxplot["medians"][0].get_ydata()[1]
            - boxplot["medians"][0].get_ydata()[0]
        ) * 3
        for i, true_val in enumerate(
            true_z, start=1
        ):  # Boxplots are indexed from 1
            ax.vlines(
                x=true_val,
                ymin=i - box_height / 2,
                ymax=i + box_height / 2,
                color="tab:orange",
                linewidth=2,
                zorder=10,
            )
    plt.savefig(os.path.join(outdir, "1d_marginals.png"))
    plt.close()


def plot_ci(
    y_obs: jnp.ndarray,
    z_posterior: jnp.ndarray,
    starccato_vae: StarccatoVAE,
    y_true: jnp.ndarray = None,
    nsamps=100,
    fname=None,
):
    y_preds = starccato_vae.generate(z=z_posterior)
    ypred_qtls = pointwise_ci(y_preds, ci=0.9)
    yrecn_qtls, yrecn_cov = None, None

    posterior_label = "Posterior"
    if y_true is not None:
        ypred_cov = coverage_probability(ypred_qtls, y_true)
        y_recon = starccato_vae.reconstruct(y_true, n_reps=nsamps)
        yrecn_qtls = pointwise_ci(y_recon, ci=0.9)
        yrecn_cov = coverage_probability(yrecn_qtls, y_true)
        posterior_label = f"Posterior (coverage: {ypred_cov:.0%})"

    plt.figure(figsize=(5, 3.5))
    plt.plot(
        y_obs, label="Observed", color="black", lw=2, zorder=-10, alpha=0.2
    )
    if y_true is not None:
        plt.plot(y_true, label="True", color="black", lw=2, zorder=0)
    ax = plt.gca()

    add_quantiles(ax, ypred_qtls, posterior_label, "tab:orange")
    if yrecn_qtls is not None:
        add_quantiles(
            ax,
            yrecn_qtls,
            f"Reconstruction (coverage: {yrecn_cov:.0%})",
            "tab:green",
        )

    ax.set_xlim(0, len(y_obs))

    plt.legend(frameon=False)
    if fname is not None:
        plt.savefig(fname)



import os

import arviz as az
import numpy as np
import pytest
import xarray as xr

from starccato_sampler.pp_test import pp_test


def generate_fake_results(
    outdir: str, num_results: int = 100, ndraw: int = 200, ndim: int = 4
):
    # Generate fake results
    np.random.seed(42)
    for i in range(num_results):
        _save_fake_res(f"{outdir}/fake_{i}", ndraw=ndraw, dim=ndim)


def _save_fake_res(
    outdir: str, nchain: int = 2, ndraw: int = 200, dim: int = 4
):
    fname = os.path.join(outdir, "inference.nc")
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    z_post = np.random.randn(nchain, ndraw, dim)
    z_true = np.random.randn(dim)
    # Convert to xarray.DataArray
    z_post_xr = xr.DataArray(
        z_post,
        dims=["chain", "draw", "dim"],  # Define dimension names
        coords={
            "chain": np.arange(nchain),
            "draw": np.arange(ndraw),
            "dim": np.arange(dim),
        },  # Coordinate labels
        name="z",  # Give it a name
    )
    idata = az.from_dict(
        posterior={"z": z_post_xr}, sample_stats={"true_latent": z_true}
    )

    assert (
        idata.posterior["z"].stack(sample=("chain", "draw")).values.size > 0
    )  # Ensure it's not empty
    assert (
        idata.sample_stats["true_latent"].to_numpy().ravel().size > 0
    )  # Ensure it's not empty
    idata.to_netcdf(fname)


def test_pp_test(tmpdir, outdir):
    dim = 4
    generate_fake_results(tmpdir, num_results=10, ndraw=20, ndim=dim)
    pp_test(
        result_regex=f"{tmpdir}/fake_*/inference.nc",
        credible_levels_fpath=f"{tmpdir}/credible_levels.npy",
        plot_fname=f"{outdir}/pp_plot.png",
        include_title=True,
        dim=dim,
    )



import jax.numpy as jnp
import jax.random as random
import numpyro
import numpyro.distributions as dist
from jax.random import PRNGKey
from numpyro.infer import MCMC, NUTS
from starccato_jax import StarccatoVAE


def _bayesian_model(
    y_obs: jnp.ndarray,
    starccato_vae: StarccatoVAE,
    rng: PRNGKey,
    beta: float,
    noise_sigma: float = 1.0,
):
    """
    Bayesian model with tempering.

    Parameters:
      y_obs   : observed data.
      vae_data: model data (e.g. containing latent_dim).
      beta    : tempering parameter; beta=1 corresponds to full posterior.
    """

    # Define priors for the latent variables
    theta = numpyro.sample(
        "z", dist.Normal(0, 1).expand([starccato_vae.latent_dim])
    )
    # Generate the signal
    y_model = starccato_vae.generate(z=theta, rng=rng)

    # sigma = numpyro.sample("sigma", dist.HalfNormal(1))  # Noise level

    # Compute the untempered log–likelihood (Assuming Gaussian noise)
    lnl = dist.Normal(y_model, noise_sigma).log_prob(y_obs).sum()

    # Temper the likelihood (the power likelihood)
    numpyro.factor("likelihood", beta * lnl)

    # Save the untempered log–likelihood (for LnZ computation).
    numpyro.deterministic("untempered_loglike", lnl)


def _run_mcmc(
    y_obs: jnp.ndarray,
    starccato_vae: StarccatoVAE,
    num_samples: int,
    num_warmup: int,
    num_chains: int,
    rng: PRNGKey,
    beta: float = 1.0,
    progress_bar: bool = False,
    noise_sigma: float = 1.0,
) -> MCMC:
    """
    Run MCMC sampling.

    Parameters:
      y_obs       : observed data.
      vae_data    : model data (e.g. containing latent_dim).
      num_samples : number of samples to draw.
      num_warmup  : number of warmup steps.
      num_chains  : number of chains.
      rng         : random number generator.
      beta        : tempering parameter; beta=1 corresponds to full posterior.
    """
    nuts_kernel = NUTS(
        lambda y_obs: _bayesian_model(
            y_obs, starccato_vae, rng, beta=beta, noise_sigma=noise_sigma
        )
    )
    mcmc = MCMC(
        nuts_kernel,
        num_warmup=num_warmup,
        num_samples=num_samples,
        num_chains=num_chains,
        progress_bar=progress_bar,
    )
    mcmc.run(rng, y_obs=y_obs)
    return mcmc



import os
from time import process_time
from typing import Dict

import arviz as az
import jax.numpy as jnp
import jax.random as random
import numpy as np
from numpyro.infer import MCMC, NUTS
from starccato_jax import StarccatoVAE
from starccato_jax.credible_intervals import coverage_probability, pointwise_ci
from tqdm.auto import tqdm

from .core import _run_mcmc
from .post_proc import _post_process


def sample(
    data: jnp.ndarray,
    model_path: str = None,
    rng_int: int = 0,
    outdir="out_mcmc",
    num_warmup=500,
    num_samples=1000,
    num_chains=1,
    noise_sigma=1.0,
    stepping_stone_lnz: bool = False,
    verbose=True,
    truth=None,
    **lnz_kwargs,
) -> MCMC:
    """
    Sample latent variables given the data.
    """
    rng = random.PRNGKey(rng_int)
    starccato_vae = StarccatoVAE(model_path)

    mcmc_kwgs = dict(
        y_obs=data,
        starccato_vae=starccato_vae,
        num_samples=num_samples,
        num_warmup=num_warmup,
        rng=rng,
        noise_sigma=noise_sigma,
    )

    t0 = process_time()
    mcmc = _run_mcmc(
        **mcmc_kwgs, beta=1.0, num_chains=num_chains, progress_bar=verbose
    )
    inf_object = az.from_numpyro(mcmc)
    inf_object.sample_stats["sampling_runtime"] = process_time() - t0

    _post_process(
        inf_object,
        data,
        truth,
        starccato_vae,
        mcmc_kwgs,
        outdir,
        stepping_stone_lnz,
        **lnz_kwargs,
    )
    return mcmc



from typing import Tuple

import jax
import matplotlib.pyplot as plt
import numpy as jnp
import numpy as np
from jax.random import PRNGKey
from jax.scipy.special import logsumexp

# from jax import numpy as jnp


def compute_stepping_stone_evidence(
    ln_likes: jnp.ndarray, betas: jnp.ndarray, outdir: str, rng: PRNGKey
) -> Tuple[float, float]:
    """
    Compute the evidence using the stepping stone approximation.

    See Patricio's paper https://arxiv.org/abs/1810.04488 and
    https://pubmed.ncbi.nlm.nih.gov/21187451/ for details.

    The uncertainty calculation is hopefully combining the evidence in each
    of the steps.

    Parameters
    ----------
    ln_likes: jnp.ndarray
        Array of log-likelihoods for each sample at each beta.
        (n_samples, n_temps-1)
    betas: jnp.ndarray
        Array of inverse temperatures.
        (n_temps,)
    outdir: str
        Output directory for plots.
    rng: PRNGKey


    Returns
    -------
    ln_z: float
        Estimate of the natural log evidence
    ln_z_err: float
        Estimate of the uncertainty in the evidence
    """

    steps, ntemps = ln_likes.shape
    if ntemps > steps:
        raise ValueError(f"Sus.. ntemps {ntemps} > steps {steps}")

    ln_z, ln_ratio = _calculate_stepping_stone(ln_likes, betas)

    # Patricio's bootstrap method to estimate the evidence uncertainty.
    ll = 50  # Block length
    repeats = 100  # Repeats
    ln_z_realisations = []
    try:
        for _ in range(repeats):
            idxs = jax.random.randint(rng, (steps - ll,), 0, steps - ll)
            ln_z_realisations.append(
                _calculate_stepping_stone(ln_likes[idxs, :], betas)[0]
            )
        ln_z_err = jnp.std(jnp.array(ln_z_realisations))
    except ValueError as e:
        print("Failed to estimate stepping stone uncertainty: ", e)
        ln_z_err = jnp.nan

    _create_stepping_stone_plot(means=ln_ratio, outdir=outdir)

    return ln_z, ln_z_err


def _calculate_stepping_stone(
    ln_likes: jnp.ndarray, betas: jnp.ndarray
) -> Tuple[float, jnp.ndarray]:
    """
    log z = Sum_k^(K−1)  Log Sum_i^(n)  L(X|θ_i, β_k-1) ** (β_k - β_k-1) - log n (k-1)
    """
    n, k = ln_likes.shape

    d_betas = betas[1:] - betas[:-1]  # beta_k - beta_{k-1}
    lnls = ln_likes[:, :-1]  # only the k-1 LnLs chains (discard
    ln_ratio = logsumexp(d_betas * lnls, axis=0) - jnp.log(n)
    return jnp.sum(ln_ratio), ln_ratio


def _create_stepping_stone_plot(means: jnp.ndarray, outdir: str):
    n_steps = len(means)

    fig, axes = plt.subplots(nrows=2, figsize=(8, 10))

    ax = axes[0]
    ax.plot(np.arange(1, n_steps + 1), means)
    ax.set_xlabel("$k$")
    ax.set_ylabel("$r_{k}$")

    ax = axes[1]
    ax.plot(np.arange(1, n_steps + 1), np.cumsum(means[::1])[::1])
    ax.set_xlabel("$k$")
    ax.set_ylabel("Cumulative $\\ln Z$")

    plt.tight_layout()
    fig.savefig(f"{outdir}/lnz_stepping_stone.png")
    plt.close()
